#include <stdlib.h>

#include "Constants.h"
#include "DataBus.h"
#include "Memory.h"

static Uint8* _memory = NULL;
static Uint8 _videoMemoryModifiedLastFrame;
static Uint8 _videoMemoryModifiedThisFrame;

static Uint8 _videoMemoryTopThirdModifiedLastFrame;
static Uint8 _videoMemoryTopThirdModifiedThisFrame;

static Uint8 _videoMemoryMiddleThirdModifiedLastFrame;
static Uint8 _videoMemoryMiddleThirdModifiedThisFrame;

static Uint8 _videoMemoryBottomThirdModifiedLastFrame;
static Uint8 _videoMemoryBottomThirdModifiedThisFrame;

void memory_mark_frame_start() {
	_videoMemoryModifiedLastFrame = _videoMemoryModifiedThisFrame;
	_videoMemoryModifiedThisFrame = 0;

	_videoMemoryTopThirdModifiedLastFrame = _videoMemoryTopThirdModifiedThisFrame;
	_videoMemoryTopThirdModifiedThisFrame = 0;

	_videoMemoryMiddleThirdModifiedLastFrame = _videoMemoryMiddleThirdModifiedThisFrame;
	_videoMemoryMiddleThirdModifiedThisFrame = 0;

	_videoMemoryBottomThirdModifiedLastFrame = _videoMemoryBottomThirdModifiedThisFrame;
	_videoMemoryBottomThirdModifiedThisFrame = 0;
}

Uint8 memory_was_video_memory_modified_last_frame() {
	return _videoMemoryModifiedLastFrame;
}

Uint8 memory_was_video_memory_modified_this_frame() {
	return _videoMemoryModifiedThisFrame;
}

void memory_start() {
	if (_memory != NULL) {
		memory_destroy();
	}
	_memory = (Uint8*)malloc(SPECTRUM_ROM_SIZE + SPECTRUM_RAM_SIZE);
	if (_memory == NULL) {
		return;
	}

	// initialize memory bytes
	for (int i = 0; i < SPECTRUM_ROM_SIZE + SPECTRUM_RAM_SIZE; i++) {
		_memory[i] = 0x00;
	}

	_videoMemoryModifiedLastFrame = 1;
	_videoMemoryModifiedThisFrame = 1;

	_videoMemoryTopThirdModifiedLastFrame = 1;
	_videoMemoryTopThirdModifiedThisFrame = 1;

	_videoMemoryMiddleThirdModifiedLastFrame = 1;
	_videoMemoryMiddleThirdModifiedThisFrame = 1;

	_videoMemoryBottomThirdModifiedLastFrame = 1;
	_videoMemoryBottomThirdModifiedThisFrame = 1;
}

void memory_destroy() {
	if (_memory != NULL) {
		free(_memory);
		_memory = NULL;
	}
}

void memory_load(Uint8* source, Uint16 offset, Uint16 size) {
	if (offset + size > SPECTRUM_ROM_SIZE + SPECTRUM_RAM_SIZE) {
		return;
	}
	SDL_memcpy(_memory + offset, source, size);
}

Uint8 memory_read8(Uint16 address) {
	Uint8 value = _memory[address & 0xffff];
	data_bus_write8(value);
	return value;
}

Uint16 memory_read16(Uint16 address) {
	Uint8 msb = _memory[(address + 1) & 0xffff];
	data_bus_write8(msb);

	return _memory[address & 0xffff] + (msb << 8);
}

void memory_write8(Uint16 address, Uint8 value) {
	data_bus_write8(value);

	if (address < SPECTRUM_ROM_SIZE) {
		return;
	}

	if (!_videoMemoryModifiedThisFrame) {
		if (address < SPECTRUM_VIDEO_RAM_START + SPECTRUM_VIDEO_RAM_LENGTH && address >= SPECTRUM_VIDEO_RAM_START) {
			_videoMemoryModifiedThisFrame = 1;
		}
	}
		
	if (!_videoMemoryTopThirdModifiedThisFrame) {
		if (address >= SPECTRUM_VIDEO_RAM_TOP_THIRD_DATA_START && address < (SPECTRUM_VIDEO_RAM_TOP_THIRD_DATA_START + SPECTRUM_VIDEO_RAM_TOP_THIRD_DATA_LENGTH)) {
			// pixel data in the top third of the screen was modified
			_videoMemoryTopThirdModifiedThisFrame = 1;
		}

		if (address >= SPECTRUM_VIDEO_RAM_TOP_THIRD_ATTRIBUTE_AREA_START && address < (SPECTRUM_VIDEO_RAM_TOP_THIRD_ATTRIBUTE_AREA_START + SPECTRUM_VIDEO_RAM_TOP_THIRD_ATTRIBUTE_AREA_DATA_LENGTH)) {
			// attribute data in the top third of the screen was modified
			_videoMemoryTopThirdModifiedThisFrame = 1;
		}
	}

	if (!_videoMemoryMiddleThirdModifiedThisFrame) {
		if (address >= SPECTRUM_VIDEO_RAM_MIDDLE_THIRD_DATA_START && address < (SPECTRUM_VIDEO_RAM_MIDDLE_THIRD_DATA_START + SPECTRUM_VIDEO_RAM_MIDDLE_THIRD_DATA_LENGTH)) {
			// pixel data in the middle third of the screen was modified
			_videoMemoryMiddleThirdModifiedThisFrame = 1;
		}

		if (address >= SPECTRUM_VIDEO_RAM_MIDDLE_THIRD_ATTRIBUTE_AREA_START && address < (SPECTRUM_VIDEO_RAM_MIDDLE_THIRD_ATTRIBUTE_AREA_START + SPECTRUM_VIDEO_RAM_MIDDLE_THIRD_ATTRIBUTE_AREA_DATA_LENGTH)) {
			// attribute data in the middle third of the screen was modified
			_videoMemoryMiddleThirdModifiedThisFrame = 1;
		}
	}

	if (!_videoMemoryBottomThirdModifiedThisFrame) {
		if (address >= SPECTRUM_VIDEO_RAM_BOTTOM_THIRD_DATA_START && address < (SPECTRUM_VIDEO_RAM_BOTTOM_THIRD_DATA_START + SPECTRUM_VIDEO_RAM_BOTTOM_THIRD_DATA_LENGTH)) {
			// pixel data in the bottom third of the screen was modified
			_videoMemoryBottomThirdModifiedThisFrame = 1;
		}

		if (address >= SPECTRUM_VIDEO_RAM_BOTTOM_THIRD_ATTRIBUTE_AREA_START && address < (SPECTRUM_VIDEO_RAM_BOTTOM_THIRD_ATTRIBUTE_AREA_START + SPECTRUM_VIDEO_RAM_BOTTOM_THIRD_ATTRIBUTE_AREA_DATA_LENGTH)) {
			// attribute data in the bottom third of the screen was modified
			_videoMemoryBottomThirdModifiedThisFrame = 1;
		}
	}
	

	_memory[address] = value;
}

void memory_write16(Uint16 address, Uint16 value) {
	memory_write8(address, (Uint8)value);			// LSB
	memory_write8(address+1, (Uint8)(value >> 8));	// MSB
}
